;;;; crunch - optimizer

(module (crunch compiler optimize) (optimize)

  (import (scheme base)
          (scheme cxr)
          (chicken base)
          (chicken format)
          (chicken plist)
          (chicken pretty-print)
          srfi-1
          matchable)

  (import (crunch compiler util)
          (crunch compiler node)
          (crunch compiler type))

  (define failed (make-parameter #f))

  (define (push! x sym prop)
    (let ((old (get sym prop '())))
      (put! sym prop (cons x old))))

  (define (pushnew! x sym prop)
    (let ((old (get sym prop '())))
      (unless (memq x old)
        (put! sym prop (cons x old)))))

  (define (merge! xs sym prop)
    (let ((old (get sym prop '())))
      (put! sym prop (lset-union eq? xs old))))

  (define (pop! x sym prop)
    (let ((old (get sym prop '())))
      (put! sym prop (delete x old))))

  (define (fail n fstr . args)
    (let ((loc (node-location n)))
      (failed #t)
      (elucidate
        (if loc
            (format "Error: ~a: ~?: ~s\n" (node-location n) fstr args
              (node-shape n))
            (format "Error: ~?: ~s\n" fstr args (node-shape n))))
      #f))

  ;; analysis pass:
  ;;
  ;; build call graph, collect free variables, mark used globals,
  ;; propagate constant vars, drop branches of known conditionals
  (define (oa n)
    (let ((seen '()))   ; all lambda id's
      (define (walk n le here)
        (case (node-class n)
          ((global-ref)
           (match (pure-value-set (node-type n))
             ((c)
              (node-class-set! n 'const)
              (node-data-set! n (list c)))
             (_ (when here
                  (pushnew! here (first (node-data n)) 'crunch:used)))))
          ((local-ref)
           (match (node-data n)
             ((a v)
              (match (pure-value-set (node-type n))
                ((c)
                 (node-class-set! n 'const)
                 (node-data-set! n (list c)))
                (_ (when (and here (not (memq a le)))
                     (put! a 'crunch:freevar-type (node-type n))
                     (pushnew! here a 'crunch:free-in)
                     (pushnew! a here 'crunch:freevars))
                   (when here (pushnew! here a 'crunch:used)))))))
          ((global-set!)
           (when here (pushnew! here (first (node-data n)) 'crunch:assigned))
           (walk (first (node-subnodes n)) le here))
          ((local-set!)
           (let ((a (first (node-data n)))
                 (val (first (node-subnodes n))))
             (unless (memq a le) (pushnew! a here 'crunch:freevars))
             (when here (pushnew! here a 'crunch:assigned))
             (walk val le here)))
          ((let)
           (match (node-data n)
             ((vs as)
              (when here
                (for-each (lambda (a) (put! a 'crunch:owner here)) as))
              (for-each (cut walk <> le here) (butlast (node-subnodes n)))
              (walk (last (node-subnodes n))
                    (append as le)
                    here))))
          ((lambda)
           (match (node-data n)
             ((id _ _ as _)
              (set! seen (lset-adjoin eq? seen id))
              (put! id 'crunch:lambda-node n)
              (for-each (lambda (a) (put! a 'crunch:owner id)) as)
              (walk (first (node-subnodes n))
                    as
                    id)
              (when (get id 'crunch:escapes)
                (let ((vs (pure-value-set (node-type (first (node-subnodes n))))))
                  (for-each
                    (lambda (vs)
                      (when (symbol? vs)
                        (put! vs 'crunch:escapes #t)))
                    vs)))
              (when here
                (let ((nfv (filter
                             (lambda (fv) (not (eq? here (get fv 'crunch:owner))))
                             (get id 'crunch:freevars '()))))
                  (merge! nfv here 'crunch:freevars)
                  (for-each
                    (lambda (fv)
                      (pushnew! here fv 'crunch:free-in))
                    nfv))))))
          ((idcall)
           (match (node-data n)
             ((id tl _)
              (when (symbol? id) ; otherwise: operator
                (unless here
                  (fail n "invalid procedure call at toplevel"))
                (push! (cons here n) id 'crunch:call-sites)
                (pushnew! here id 'crunch:callers)
                (pushnew! id here 'crunch:callees))
              (for-each (cut walk <> le here) (node-subnodes n)))))
          ((if)
           (match (node-subnodes n)
             ((t c a)
              (match (pure-value-set (node-type t))
                ((#f)
                 (explain n "dropping constant conditional test (false)")
                 (node-class-set! n 'begin)
                 (node-subnodes-set! n (list t a)))
                ((val1 vals ...)
                 (unless (memq #f (cons val1 vals))
                   (explain n "dropping constant conditional test (non-false)")
                   (node-class-set! n 'begin)
                   (node-subnodes-set! n (list t c))))
                (_ #f))
              (for-each (cut walk <> le here) (node-subnodes n)))))
          ((call)
           (unless here
             (fail n "invalid procedure call at toplevel"))
           (for-each (cut walk <> le here) (node-subnodes n)))
          ;; unspecified, const, begin, call, dispatch, c-lambda, c-declare, c-external
          (else (for-each (cut walk <> le here) (node-subnodes n)))))
      (dribble "gathering information")
      (walk n '() #f)
      ;; mark lambda-ids in result type-sets of ecaping procedures
      ;; as escaping
      (let rec ((ids (filter (cut get <> 'crunch:escapes) seen))
                (m (gensym)))
        (cond ((null? ids))
              ((not (eq? m (get (car ids) 'crunch:seen)))
                (put! (car ids) 'crunch:seen m)
                (put! (car ids) 'crunch:escapes #t)
                (let ((vs (pure-value-set
                            (procedure-result-type
                              (node-type (get (car ids) 'crunch:lambda-node))))))
                  (rec (filter symbol? vs) m)))
              (else (rec (cdr ids) m))))
      n))

  ;; globalization pass:
  ;;
  ;; identify let-bound variables that occur at toplevel or in a non-escaping
  ;; lambda that has only a single call site.
  (define (og n)
    (define (walk n si)  ; si = single instance container lambdas
      (case (node-class n)
        ((lambda)
         (let ((id (first (node-data n))))
           (walk (first (node-subnodes n))
                 (and si
                      (not (get id 'crunch:escapes))
                      (= 1 (length (get id 'crunch:call-sites '())))))))
        ((let)
         (match (node-data n)
           ((vs as)
            (cond (si
                    (node-class-set! n 'begin)
                    (node-data-set! n '())
                    (let ((loc (node-location n)))
                      (let loop ((vs vs) (as as) (ns (node-subnodes n)))
                        (cond ((null? vs) (walk (car ns) #t))
                              (else
                                (let ((a (car as))
                                      (n (car ns))
                                      (v (car vs)))
                                  (unless (eq? v a)
                                    (explain n "globalized variable `~a'" v))
                                  (walk n #t)
                                  (set-car!
                                    ns
                                    (make-node loc
                                               'global-set!
                                               (list a #f a)
                                               (list n)
                                               (node-type n)))
                                  (put! a 'crunch:globalized #t)
                                  (for-each
                                    (lambda (id)
                                      (pop! a id 'crunch:freevars))
                                    (get a 'crunch:free-in '()))
                                  (put! a 'crunch:free-in '())
                                  (loop (cdr vs) (cdr as) (cdr ns))))))))
                  (else (for-each (cut walk <> si) (node-subnodes n)))))))
        ((local-ref)
         (let ((a (first (node-data n))))
           (when (get a 'crunch:globalized)
             (node-class-set! n 'global-ref)
             (node-data-set! n (list a #f a)))))
        ((local-set!)
         (let ((a (first (node-data n)))
               (val (first (node-subnodes n))))
           (when (get a 'crunch:globalized)
             (node-class-set! n 'global-set!)
             (node-data-set! n (list a #f a)))
           (walk val si)))
        (else (for-each (cut walk <> si) (node-subnodes n)))))
    (dribble "globalization pass")
    (walk n #t)
    n)

  ;; integration pass:
  ;;
  ;; inline procedures called only once, loop-detection
  (define (oi n)
    (define (integrate n id inid asloop)
      (explain n "integrating call to `~a' in `~a'"
                (lambda-name id) (lambda-name inid))
      (let* ((ln (get id 'crunch:lambda-node))
             (args (node-subnodes n))
             (pt (get id 'crunch:lambda))
             (argt (procedure-argument-types pt))
             (lbody (first (node-subnodes ln))))
       (match (node-data ln)
          ((_ _ vs as _)
            (node-class-set! n 'let)
            (node-data-set! n (list vs as))
            (node-subnodes-set!
              n
              (append (integration-arguments args vs argt)
                      (list lbody)))
            (node-type-set! n (procedure-result-type pt))
            (merge! (delete id (get id 'crunch:callees '()) eq?)
                    inid 'crunch:callees)
            (put! id 'crunch:callees '())
            (put! id 'crunch:call-sites '())
            (pop! inid id 'crunch:callers)
            (pop! id inid 'crunch:callees)
            (for-each
              (lambda (fv)
                (pop! id fv 'crunch:free-in))
              (get id 'crunch:freevars '()))
            (put! id 'crunch:freevars '())
            (put! id 'crunch:lambda #f)
            (put! id 'crunch:lambda-node #f)
            (put! id 'crunch:integrated inid)
            (adjust-uses/assigns id inid n)
            (cond (asloop
                    (explain n "converting integrated procedure `~a' into loop"
                             (lambda-name id))
                    (let ((l1 (gensym 'l))
                          (lbody2 (clone-node lbody)))
                      (adjust-call-sites lbody lbody2 inid)
                      (node-class-set! lbody 'labels)
                      (node-data-set! lbody (list (list l1)))
                      (node-subnodes-set!
                        lbody
                        (list (loopify lbody2 id l1 as vs argt)))))
                  (else (adjust-call-sites n n inid)))
            n))))
    (define (adjust-uses/assigns oldid newid n)
      (case (node-class n)
        ((local-ref global-ref)
          (pop! oldid (first (node-data n)) 'crunch:used)
          (pushnew! newid (first (node-data n)) 'crunch:used))
        ((local-set! global-set!)
          (pop! oldid (first (node-data n)) 'crunch:assigned)
          (pushnew! newid (first (node-data n)) 'crunch:assigned)))
      (for-each (cut adjust-uses/assigns oldid newid <>)
        (node-subnodes n)))
    (define (adjust-call-sites n n2 inid)
      (for-each (cut adjust-call-sites <> n2 inid) (node-subnodes n))
      (case (node-class n)
        ((idcall)
         (match (node-data n)
           ((id tl _)
            (when (symbol? id) ; not operator
              (let ((a (find (lambda (a)
                               (and (eq? id (car a))
                                    (eq? n (cdr a))))
                             (get id 'crunch:call-sites '()))))
                (when (pair? a)
                  (set-car! a inid)
                  (set-cdr! a n2)))))))))
    (define (integration-arguments args vs argt)
      (let rec ((args args) (vs vs) (argt argt))
        (cond ((null? vs) '())
              ((eq? '#!optional (car argt))
               (rec args vs (cdr argt)))
              ((null? args)
               (cons (make-node (node-location n)
                                'const
                                (list (default-init (car argt)))
                                '()
                                (car argt))
                     (rec '() (cdr vs) (cdr argt))))
              (else (cons (car args) (rec (cdr args) (cdr vs) (cdr argt)))))))
    (define (loopify n id entry as vs argt)
      (case (node-class n)
        ((idcall)
          (node-subnodes-set!
            n
            (map (cut loopify <> id entry as vs argt) (node-subnodes n)))
          (when (and (eq? id (first (node-data n)))
                     (second (node-data n)))
            (explain n "converting tail-recursive procedure `~a' into loop"
                     (lambda-name id))
            (pop! id id 'crunch:callers)
            (put! id 'crunch:call-sites
                  (remove
                    (match-lambda
                      ((id2 . n2)
                        (and (eq? id2 id) (eq? n n2)))
                      (_ #f))
                    (get id 'crunch:call-sites '())))
            (let* ((args (integration-arguments (node-subnodes n) vs argt))
                   (tmps+args
                     (filter-map
                        (lambda (a v arg)
                          (and (not (eq? (node-class arg) 'unspecified))
                               (or (not (eq? (node-class arg) 'const))
                                   (not (symbol? (first (node-data arg)))))
                               (list (gensym 'la) a v arg)))
                        as vs args))
                   (tmps (map first tmps+args))
                   (loc (node-location n)))
              (node-class-set! n 'let)
              (node-data-set! n (list tmps tmps))
              (node-subnodes-set!
                n
                (append
                  (map fourth tmps+args)
                  (list
                    (make-node
                      loc 'begin '()
                      (append
                        (map (match-lambda
                               ((tmp a v arg)
                                 (put! tmp 'crunch:owner id)
                                 (pushnew! id tmp 'crunch:used)
                                 (pushnew! id a 'crunch:assigned)
                                 (make-node
                                   loc 'local-set!
                                   (list a v (node-type arg))
                                   (list
                                     (make-node
                                       loc 'local-ref (list tmp tmp)
                                       '()
                                       (node-type arg)))
                                   (void-type))))
                             tmps+args)
                        (list (make-node
                                loc 'goto (list entry) '()
                                (void-type))))
                      (void-type)))))))
          n)
        (else
          (node-subnodes-set!
            n
            (map (cut loopify <> id entry as vs argt) (node-subnodes n)))
          n)))
    (define (walk n e here)
      (case (node-class n)
        ((idcall)
          (node-subnodes-set! n (map (cut walk <> e here) (node-subnodes n)))
          (let* ((id (first (node-data n)))
                 (pcall (symbol? id)) ; not (operator _)
                 (sites (and pcall (get id 'crunch:call-sites))))
            (if (and sites
                     (pair? sites)
                     (get id 'crunch:lambda-node)
                     (not (get id 'crunch:escapes))
                     (not (get id 'crunch:primitive))
                     (lset<= eq? (get id 'crunch:freevars '()) e))
                (let loop ((sites sites)
                           (this #f)
                           (loops #f))
                  (match sites
                    (((where . call) . more)
                      (let repeat ((where where))
                        (cond ((eq? n call) (loop more #t loops))
                              ((get where 'crunch:integrated) => repeat)
                              ((and (eq? where id)
                                    (second (node-data call))) ; tailcall?
                               (loop more this #t))
                              (else
                                n))))
                     (()
                       (if this
                           (integrate n id here loops)
                           n))))
                n)))
        ((lambda)
          (match (node-data n)
            ((id _ vs as _)
              (when (any (match-lambda
                           ((where . call)
                             (and (eq? id where)
                                  (second (node-data call))))) ; tailcall?
                         (get id 'crunch:call-sites '()))
                (let ((entry (gensym 'l))
                      (pt (get id 'crunch:lambda)))
                  (node-subnodes-set!
                    n
                    (list
                      (make-node (node-location n)
                                 'labels (list (list entry))
                                 (list
                                   (loopify (first (node-subnodes n))
                                            id entry as vs
                                            (procedure-argument-types pt)))
                                 (procedure-result-type pt))))))
              (node-subnodes-set!
                n
                (list (walk (first (node-subnodes n)) (append as e) id)))
              n)))
        ((let)
          (node-subnodes-set!
            n
            (append
              (map (cut walk <> e here) (butlast (node-subnodes n)))
              (list (walk (last (node-subnodes n))
                          (append (second (node-data n)) e)
                          here))))
          n)
        (else
          (node-subnodes-set!
            n
            (map (cut walk <> e here) (node-subnodes n)))
          n)))
    (dribble "integration pass")
    (walk n '() #f))

  ;; lambda lifting pass:
  ;;
  ;; eliminate free variables and perform boxing
  (define (ol n)
    (define (findfree vars aliases)
      (filter-map
        (lambda (v a)
          (and (pair? (get a 'crunch:free-in '()))
               (list a v (pair? (get a 'crunch:assigned '())) a)))
        vars aliases))
    (define (real-lambda-id id)
      (cond ((get id 'crunch:integrated) => real-lambda-id)
            (else id)))
    (define (forward n free here)
      (case (node-class n)
        ((local-ref)
          (cond ((assq (first (node-data n)) free) =>
                  (match-lambda
                    ((a v b a2)
                      (unless (eq? a a2)
                        (pushnew! here a2 'crunch:used))
                      (if b
                          (make-node (node-location n)
                                     'box-ref '()
                                     (list (make-node (node-location n)
                                                      'local-ref
                                                      (list a2 v)
                                                      '()
                                                      (node-type n)))
                                     (node-type n))
                          (begin
                            (node-data-set! n (list a2 v))
                            n)))))
                (else n)))
        ((local-set!)
          (let ((val (first (node-subnodes n))))
            (cond ((assq (first (node-data n)) free) =>
                   (match-lambda
                     ((a v b a2)
                      (cond (b
                              (pushnew! here a2 'crunch:used)
                              (node-class-set! n 'box-set!)
                              (node-data-set! n '())
                              (node-subnodes-set!
                                n
                                (list (make-node (node-location n)
                                                 'local-ref
                                                 (list a2 v)
                                                 '()
                                                 (make-typevar
                                                   `(box ,(node-type val))
                                                   '()))
                                      (forward val free here))))
                            (else
                              (pushnew! a2 here 'crunch:assigned)
                              (set-car! (node-data n) a2)
                              (node-subnodes-set!
                                n
                                (list (forward val free here))))))))
                  (else (node-subnodes-set! n (list (forward val free here)))))
            n))
        ((lambda)
          (match (node-data n)
            ((id argc vs as w)
             (unless (get id 'crunch:integrated)
               (let* ((fv (get id 'crunch:freevars '()))
                      (free2 (filter-map
                               (match-lambda
                                 ((a v b? a2)
                                  (and (memq a fv)
                                       (list a v b? (gensym a)))))
                               free))
                      (body (first (node-subnodes n))))
                 ;; set ownership of aliases
                 (for-each
                   (lambda (f)
                     (put! (fourth f) 'crunch:owner id))
                   free2)
                 (node-data-set!
                   n
                   (list id
                         (+ argc (length free2))
                         (append (map second free2) vs)
                         (append (map fourth free2) as)
                         w))
                 ;; add freevars to all call-sites
                 (for-each
                   (match-lambda
                     ((site . _)
                      (for-each
                        (lambda (v)
                          (let ((site (real-lambda-id site))
                                (owner (real-lambda-id (get v 'crunch:owner))))
                            (unless (eq? site owner)
                              (pushnew! v site 'crunch:freevars)
                              (pushnew! site v 'crunch:free-in))))
                       fv)))
                   (get id 'crunch:call-sites '()))
                 ;; addfree is the same as freevars, but freevars must be
                 ;; eventually cleared so we can check at the end for lambdas t
                 ;; hat still have non-optimizable freevars.
                 (put! id 'crunch:addfree fv)
                 (node-subnodes-set!
                   n
                   (list (forward body free2 id)))
                 (match (typevar-type (node-type n))
                   (('procedure argts rts)
                    (type-copy!
                      (node-type n)
                      (make-typevar
                        `(procedure ,(append (map (lambda (fv)
                                                    (let ((t1 (get (first fv)
                                                                   'crunch:freevar-type)))
                                                      (if (get (first fv) 'crunch:boxed)
                                                          (make-typevar `(box ,t1) '())
                                                          t1)))
                                               free2)
                                             argts)
                                    ,rts)
                        (list id)))))
                 (for-each
                   (lambda (f)
                     (pop! id (first f) 'crunch:free-in)
                     (pop! (first f) id 'crunch:freevars))
                   free2)))
             n)))
        ((idcall)
          (let* ((id (first (node-data n)))
                 (fv (and (symbol? id)
                          (get id 'crunch:addfree)))
                 (add (filter-map
                        (match-lambda
                          ((a v b? a2)
                            (and (and fv (memq a fv))
                                 (begin
                                   (pushnew! here a2 'crunch:used)
                                   #t)
                                 (make-node (node-location n)
                                            'local-ref
                                            (list a2 v)
                                            '()
                                            #f))))
                        free)))
            (node-subnodes-set!
              n
              (append add (map (cut forward <> free here) (node-subnodes n))))
            n))
        ;; global-ref, global-set!, if, call, let, dispatch, const, begin,
        ;; variant, goto, labels, unspecified
        (else
          (node-subnodes-set! n (map (cut forward <> free here) (node-subnodes n)))
          n)))
    ;; re-mark ownership (may have changed after integration)
    (define (fix-owners n here)
      (case (node-class n)
        ((lambda)
          (match (node-data n)
            ((id _ _ as _)
              (unless (get id 'crunch:integrated)
                (for-each (cut put! id <> 'crunch:owner) as)
                (fix-owners (first (node-subnodes n)) id)))))
        ((let)
          (for-each (cut put! here <> 'crunch:owner)
            (second (node-data n)))
          (for-each (cut fix-owners <> here) (node-subnodes n)))
        (else
          (for-each (cut fix-owners <> here) (node-subnodes n)))))
    ;; walk and forward variable bindings that are free in inner nodes
    (define (walk n here)
      (case (node-class n)
        ((lambda)
          (match (node-data n)
            ((id argc vs as _)
             (unless (get id 'crunch:integrated)
               (let* ((free (findfree vs as))
                      (boxed (filter third free))
                      (body (first (node-subnodes n)))
                      (loc (node-location body)))
                 (for-each
                   (lambda (bx)
                     (put! (first bx) 'crunch:boxed #t)
                     (put! (first bx) 'crunch:owner id))
                   boxed)
                 (node-subnodes-set!
                   n
                   (if (pair? free)
                       (list (walk (forward body free here) id))
                       (list (walk body id))))))
             n)))
        ((let)
          (match (node-data n)
            ((vs as)
              (let* ((free (findfree vs as))
                     (boxed (filter third free))
                     (body (last (node-subnodes n)))
                     (loc (node-location body)))
                (for-each
                  (lambda (bx)
                    (put! (first bx) 'crunch:boxed #t))
                  boxed)
                (node-data-set!
                  n
                  (list vs
                        (map (lambda (a)
                               (let ((ba (assq a free)))
                                 (cond (ba
                                         (put! (fourth ba) 'crunch:owner here)
                                         (fourth ba))
                                       (else a))))
                             as)))
                (node-subnodes-set!
                  n
                  (map (cut walk <> here)
                       (append (butlast (node-subnodes n))
                               (list (forward body free here)))))
                n))))
        ;; global-ref, global-set!, local-ref, global-ref, if, call, let,
        ;; dispatch, const, begin, variant, goto, idcall, labels, unspecified
        (else
          (node-subnodes-set! n (map (cut walk <> here) (node-subnodes n)))
          n)))
    (dribble "eliminating free variables")
    (fix-owners n #f)
    (walk n #f))

  ;; DCE pass:
  ;;
  ;; dead code elimination, drop unused let-bindings and global
  ;; assignments, drop unused parameters, drop unused labels-blocks,
  ;; merge let nodes and replace initial binding with assignment (if needed),
  ;; straighten inner begins in "let", "if" and "set!"
  (define (od n)
    (let ((globalized '())
          (let-bound '())) ; ((<var> <alias> <init>) ...)
      ;; "push" assignment into final tail-positions
      (define (push-assignment loc class data n)
        (let walk ((n n))
          (case (node-class n)
            ((begin let)
              (walk (last (node-subnodes n)))
              n)
            ((goto local-set! global-set! unspecified)
              ;; does not return or has no result
              n)
            ((labels)
              (for-each walk (node-subnodes n))
              n)
            ((if)
              (match (node-subnodes n)
                ((_ x y)
                  (walk x)
                  (walk y)
                  n)))
            ((switch dispatch)
              (for-each walk (cdr (node-subnodes n)))
              n)
            (else
              (replace-node!
                n
                (make-node
                  loc class data
                  (list (clone-node n))
                  (void-type)))
              n))))
      ;; convert cascade of "if" nodes to %switch
      (define (switchify n)
        (define (fuse val n)
          (let ((other (last (node-subnodes n))))
            (or (and (eq? 'switch (node-class other))
                     (match (node-subnodes other)
                       ((val2 case1 cases ...)
                        (and (eq? (node-class val) (node-class val2))
                             (equal? (node-data val) (node-data val2))
                             (fuse
                               val
                               (make-node
                                 (node-location n)
                                 'switch
                                 (list (append (first (node-data n))
                                               (first (node-data other))))
                                 (cons val
                                       (append (butlast (cdr (node-subnodes n)))
                                               (cons case1 cases)))
                                 (node-type n)))))))
                n)))
        (or (and (eq? 'if (node-class n))
                 (let ((tn (first (node-subnodes n))))
                   (and (eq? 'idcall (node-class tn))
                        (equal? '(operator ==) (first (node-data tn)))
                        (match (node-subnodes tn)
                          ((x1 x2)
                           (cond ((and (eq? 'local-ref (node-class x1))
                                       (eq? 'const (node-class x2))
                                       (integer? (first (node-data x2))))
                                  (fuse
                                    x1
                                    (make-node
                                      (node-location n)
                                      'switch
                                      (list (list (first (node-data x2))))
                                      (cons x1 (cdr (node-subnodes n)))
                                      (node-type n))))
                                 ((and (eq? 'local-ref (node-class x2))
                                       (eq? 'const (node-class x1))
                                       (integer? (first (node-data x1))))
                                  (fuse
                                    x2
                                    (make-node
                                      (node-location n)
                                      'switch
                                      (list (list (first (node-data x1))))
                                      (cons x2 (cdr (node-subnodes n)))
                                      (node-type n))))
                                 (else n)))))))
            n))
      ;; walk and simplify
      (define (walk n top)
        (define (obsolete-value-set? tv)
          (let ((vs (pure-value-set tv)))
            (and (pair? vs)
                 (every (lambda (id)
                          (and (symbol? id)
                               (get id 'crunch:integrated)))
                        vs))))
        (define (nonvalue-ref? n)
          (and (memq (node-class n) '(local-ref global-ref))
               (eq? 'void (typevar-type (node-type n)))))
        (case (node-class n)
          ((begin)
           (let loop ((ns (node-subnodes n)) (seq '()))
             (match ns
               ((n1)
                 (let ((n2 (walk n1 top)))
                   (when (nonvalue-ref? n2)
                     (void-node! n2))
                   (if (null? seq)
                       (walk n2 top)
                       (make-node (node-location n)
                                  'begin '()
                                  (reverse (cons n2 seq))
                                  (node-type n2)))))
               ((n1 n2 more ...)
                (let ((n1 (walk n1 top)))
                  (cond ((nonvalue-ref? n1) (loop (cdr ns) seq))
                        ((ineffective? n1) (loop (cdr ns) seq))
                        ((eq? (node-class n1) 'begin)
                         (loop (append (node-subnodes n1) (cdr ns)) seq))
                        (else (loop (cdr ns) (cons n1 seq)))))))))
          ((let)
           (match (node-data n)
             ((vs as)
              (let* ((prefix '())
                     (b (walk (last (node-subnodes n)) top))
                     (bs (filter-map
                           (lambda (v a nb)
                             (cond ((and (ineffective? nb)
                                         (or (null? (get a 'crunch:used '()))
                                             (obsolete-value-set? (node-type nb))
                                             (eq? 'void
                                                  (typevar-type (node-type nb)))))
                                    #f)
                                   ((pair? (get a 'crunch:free-in '()))
                                    (fail n "local variable `~a' is free" v))
                                   (else
                                     (let repeat ((nb nb))
                                       (cond ((eq? 'begin (node-class nb))
                                              (set! prefix
                                                (append prefix
                                                        (butlast (node-subnodes nb))))
                                              (repeat (last (node-subnodes nb))))
                                             (else
                                               (let ((tv (node-type nb)))
                                                 (set! let-bound
                                                   (cons (list v a
                                                               (make-node
                                                                     (node-location nb)
                                                                     'const
                                                                     (list (default-init tv))
                                                                     '()
                                                                     tv))
                                                         let-bound))
                                                 (list v a nb))))))))
                           vs as (map (cut walk <> top)
                                   (butlast (node-subnodes n))))))
                (if (null? bs)
                    b
                    (begin
                      (node-class-set! n 'begin)
                      (node-data-set! n '())
                      (node-subnodes-set!
                        n
                        (append
                          prefix
                          (let rec ((bs bs))
                            (match bs
                              (() (list b))
                              (((v a n) bs ...)
                                (cond ((get a 'crunch:boxed)
                                         (cons
                                           (make-node
                                             (node-location n)
                                               'box-set! '()
                                               (list
                                                 (make-node
                                                   (node-location n)
                                                   'local-ref
                                                   (list a v)
                                                   '()
                                                   (node-type n))
                                                 n)
                                               (void-type))
                                           (rec bs)))
                                      ((eq? 'unspecified (node-class n))
                                         (rec bs))
                                      ((and (uninitialized-type? (node-type n))
                                            (eq? 'const (node-class n)))
                                         (rec bs))
                                      ;; duplicates test above...
                                      ((or (null? (get a 'crunch:used))
                                      	   (eq? 'void (typevar-type (node-type n)))
                                           (obsolete-value-set? (node-type n)))
                                         ;; binding value has effect but is not
                                         ;; used.
                                         (cons
                                           (walk n top)
                                           (rec bs)))
                                      (else
                                         (cons
                                           (walk
                                            (make-node
                                             (node-location n)
                                             'local-set!
                                             (list a v (node-type n))
                                             (list n)
                                             (void-type))
                                           top)
                                           (rec bs)))))))))
                      n))))))
          ((local-set!)
            (match (node-data n)
              ((a v _)
               (let ((vn (walk (first (node-subnodes n)) top)))
                 (cond ((or (null? (get a 'crunch:used '()))
                            (obsolete-value-set? (node-type vn)))
                        (cond ((ineffective? vn) (void-node! n))
                              (else vn)))
                       ((and (eq? 'local-ref (node-class vn))
                             (eq? a (first (node-data vn))))
                         ;; self-assignment
                         (void-node! n))
                       ((memq (node-class vn) '(begin labels let if switch dispatch))
                         (walk
                           (push-assignment
                             (node-location n)
                             'local-set! (node-data n)
                             vn)
                           top))
                       ((eq? 'unspecified (node-class vn)) (void-node! n))
                       (else
                         (node-subnodes-set! n (list vn))
                         n))))))
          ((global-set!)
            (match (node-data n)
              ((var exp ln)
               (let ((vn (walk (first (node-subnodes n)) top)))
                 (define (mark-globalized)
                   (when (get var 'crunch:globalized)
                     (unless (assq var globalized)
                       (set! globalized
                         (alist-cons var (node-type vn) globalized)))))
                 (cond ((and (not exp)
                             (null? (get var 'crunch:used '())))
                        (explain n "dropping assignment to unused global variable `~a'" var)
                        (cond ((ineffective? vn) (void-node! n))
                              (else vn)))
                       ((memq (node-class vn) '(begin labels let if switch))
                         (walk
                           (push-assignment
                             (node-location n)
                             'global-set! (node-data n)
                             vn)
                           top))
                       ((and (eq? 'unspecified (node-class vn))
                             (get var 'crunch:globalized))
                         (void-node! n)
                         (mark-globalized)
                         n)
                       (top
                         (let ((vn (walk vn #t)))
                           (cond ((and (eq? 'global-ref (node-class vn))
                                       (get (first (node-data vn))
                                            'crunch:initial-value))
                                  => (lambda (init)
                                       (put! (first (node-data n))
                                             'crunch:initial-value init)
                                       (void-node! n)))
                                  ((not (memq (node-class vn)
                                          '(unspecified lambda variant const
                                            c-external c-lambda)))
                                  (fail n "disallowed toplevel assignment"))
                                 (else
                                   (node-subnodes-set! n (list vn))
                                   n))))
                       (else
                         (mark-globalized)
                         (node-subnodes-set! n (list (walk vn top)))
                         n))))))
          ((lambda)
            (match (node-data n)
              ((id argc vs as _)
               (cond ((get id 'crunch:integrated)
                      (let ((body (first (node-subnodes n))))
                        (node-subnodes-set!
                          n
                          (list (make-node (node-location body)
                                           'const
                                           (list (default-init (node-type body)))
                                           '()
                                           (node-type body))))
                        n))
                     (else
                       (unless (get id 'crunch:escapes)
                         (let loop ((vs vs)
                                    (as as)
                                    (dropped '())
                                    (argts (procedure-argument-types (node-type n)))
                                    (vsnew '())
                                    (asnew '())
                                    (atsnew '())
                                    (p 0)
                                    (i 0)
                                    (opt #f))
                           (cond ((and (pair? argts)
                                       (eq? '#!optional (car argts)))
                                   (loop vs as dropped (cdr argts)
                                         vsnew asnew
                                         (cons '#!optional atsnew)
                                         p i #t))
                                 ((null? vs)
                                  (when (pair? dropped)
                                    (explain n "removed unused parameters from procedure ~a: ~s"
                                             (lambda-name id) dropped)
                                    (let* ((tv (node-type n))
                                           (rt (procedure-result-type tv))
                                           (d (node-data n)))
                                      (typevar-type-set!
                                        tv
                                        `(procedure ,(reverse atsnew) ,rt))
                                      (set-car! (cdr d) i)
                                      (set-car! (cddr d) (reverse vsnew))
                                      (set-car! (cdddr d) (reverse asnew))))
                                  n)
                                 ((and (not opt)
                                       (null? (get (car as) 'crunch:used '())))
                                   (let ((cs (get id 'crunch:call-sites '())))
                                     (for-each
                                      (lambda (a)
                                         (drop-arg! (cdr a) p))
                                     cs))
                                  (loop (cdr vs) (cdr as) (cons (car vs) dropped)
                                        (cdr argts)
                                        vsnew asnew atsnew p i opt))
                                 (else (loop (cdr vs) (cdr as) dropped
                                             (cdr argts)
                                             (cons (car vs) vsnew)
                                             (cons (car as) asnew)
                                             (cons (car argts) atsnew)
                                             (add1 p)
                                             (if (< i argc) (add1 i) i)
                                             opt)))))
                       (fluid-let ((let-bound '()))
                         (let ((body (walk (first (node-subnodes n)) #f)))
                           (node-subnodes-set!
                             n
                             (list (if (null? let-bound)
                                       body
                                       (make-node (node-location n)
                                                  'let
                                                  (list (map first let-bound)
                                                        (map second let-bound))
                                                  (append (map third let-bound)
                                                          (list body))
                                                  (node-type body)))))))
                       n)))))
          ((variant)
            (let ((lns (filter-map
                         (lambda (ln)
                           (assert (eq? 'lambda (node-class ln)))
                           ;; not integrated?
                           (and (not (get (first (node-data ln))
                                          'crunch:integrated))
                                (walk ln #t)))
                         (node-subnodes n))))
              (cond ((null? lns)
                     (make-node (node-location n) 'unspecified '() '()
                                (void-type)))
                    ((null? (cdr lns)) (car lns))
                    (else
                      (node-subnodes-set! n lns)
                      n))))
          ((goto)
            (put! (first (node-data n)) 'crunch:seen 1)
            n)
          ((labels)
            (let loop ((ns (cdr (node-subnodes n)))
                       (nsnew (list (walk (first (node-subnodes n)) top)))
                       (ls (cdr (first (node-data n))))
                       (lsnew (list (first (first (node-data n))))))
              (cond ((null? ns)
                      (node-subnodes-set! n (reverse nsnew))
                      (node-data-set! n (list (reverse lsnew)))
                      n)
                    ((eq? 1 (get (first ls) 'crunch:seen))
                      (loop (cdr ns)
                            (cons (walk (car ns) top) nsnew)
                            (cdr ls)
                            (cons (car ls) lsnew)))
                    (else
                      (loop (cdr ns) nsnew (cdr ls) lsnew)))))
          ((if)
            (let* ((ns (map (cut walk <> top) (node-subnodes n)))
                   (tn (first ns)))
              (if (eq? (node-class tn) 'begin)
                  (let ((bns (node-subnodes tn)))
                    (make-node (node-location n)
                               'begin '()
                               (append (butlast bns)
                                       (list (switchify
                                               (make-node
                                                 (node-location n)
                                                 'if '()
                                                 (cons (last bns) (cdr ns))
                                                 (node-type n)))))
                               (node-type n)))
                  (let ((ns (map (cut walk <> top) ns)))
                    (node-subnodes-set! n ns)
                    (switchify n)))))
          ;; unspecified, const, local-ref, global-ref, call, idcall, dispatch, box-set!
          ;; box-ref, c-lambda, c-external, c-declare
          (else
            (let ((ns (map (cut walk <> top) (node-subnodes n))))
              (node-subnodes-set! n ns)
              n))))
      (dribble "dead code elimination")
      (let ((n (walk n #t)))
        (when (failed) ((node-dump-hook) n "ERROR: dead code elimination failed"))
        (and (not (failed))
             (if (null? globalized)
                 n
                 (make-node
                   #f 'begin '()
                   (append
                     (map
                       (match-lambda
                         ((var . tv)
                          (make-node
                            #f 'global-set! (list var #f var)
                            (list
                              ;; note node-type: we need it to generate
                              ;; the type for the declaration
                              (let ((init (get var 'crunch:initial-value)))
                                (if init
                                    (make-node #f 'const (list n) '() tv)
                                    (make-node #f 'unspecified '() '() tv))))
                            (void-type))))
                       globalized)
                     (list n))
                   (void-type)))))))

  (define (lambda-name id)
    (get id 'crunch:external-name id))

  (define (explain n fstr . args)
    (when (memq 'explain (or (verbose) '()))
      (elucidate (format "~a: ~?\n" (or (node-location n) "<toplevel>")
                   fstr args))))

  (define (show n)
    (pp (node-shape n #f))
    n)

  (define (optimize n)
    (parameterize ((failed #f))
      ((o od ol oi og oa) n)))

)
