;;; crunch - type-specifiers and variables
;
; Types:
;
;   boolean integer float complex string char
;   (procedure (T ... [#!optional T ...]) T)
;   (pointer T)
;   pointer
;   (numvector S)
;   (struct NAME)
;   (union NAME)
;   (typename NAME)
;   (vector T)
;   port
;   (box T)		introduced during optimization (lambda lifting)
;   (values T)		for rest-list values
;   void
;   noreturn
;   ?			(unknown)
;   #f              	(no type - variant)
;
; Values in value-sets:
;
;   <lambda-id>		(symbol)
;   <boolean>
;   <number>
;   <string>
;   <character>
;   ?			(unknown)


(module (crunch compiler type) *

  (import (scheme base)
          (scheme cxr)
          srfi-1
          matchable
          (chicken base)
          (chicken plist)
          (chicken string)
          (chicken memory representation)
          (chicken number-vector))

  (import (crunch compiler util))

  (define idcount 0)

  (define-record-type typevar
    (%make-typevar id type value-set)
    typevar?
    (id typevar-id)
    (type typevar-type typevar-type-set!)
    (value-set typevar-value-set typevar-value-set-set!))

  (define (make-typevar t vs)
    (set! idcount (add1 idcount))
    (%make-typevar idcount t vs))

  (define (value-set x)
    (cond ((number? x) (list x))
          ((string? x) (list x))
          ((boolean? x) (list x))
          ((vector? x) '())
          (else '(?))))

  (define (typename t)
    (if (eq? t '#!optional)
        t
        (let ((ts (typevar-type t)))
          (match ts
            (#f #f)
            (('procedure at rt)
             `(procedure ,(map typename at) ,(typename rt)))
            (('pointer t)
             `(pointer ,(typename t)))
            (('vector t)
             `(vector ,(typename t)))
            (('values t)
             `(values ,(typename t)))
            (_ ts)))))

  (define (typespec s #!optional id)
    (match s
      ((? typevar?) s)
      ((or '? 'integer 'complex 'char 'boolean 'float 'pointer 'port 'void 'noreturn
           'string 'scheme-object)
        (make-typevar s '()))
      (('procedure ats rt)
        (make-typevar
          `(procedure ,(map (lambda (s)
                              (if (eq? s '#!optional)
                                  s
                                  (typespec s)))
                            ats)
             ,(typespec rt))
          (if id (list id) '())))
      (((or '* 'pointer) t)
        (make-typevar
          `(pointer ,(typespec t))
          '()))
      (((or 'vector 'values) t)
        (make-typevar
          (list (car s) (typespec t))
          '()))
      ('cplxnum (typespec 'complex))
      ((or 'true 'false) (typespec 'boolean))
      ((or 'u8vector 'bytevector) (typespec '(numvector u8)))
      ('s8vector (typespec '(numvector s8)))
      ('u16vector (typespec '(numvector u16)))
      ('s16vector (typespec '(numvector s16)))
      ('u32vector (typespec '(numvector u32)))
      ('s32vector (typespec '(numvector s32)))
      ('u64vector (typespec '(numvector u64)))
      ('s64vector (typespec '(numvector s64)))
      ('f32vector (typespec '(numvector f32)))
      ('f64vector (typespec '(numvector f64)))
      ('c64vector (typespec '(numvector c64)))
      ('c128vector (typespec '(numvector c128)))
      (('numvector t)
        (if (memq t '(u8 s8 u16 s16 u32 s32 u64 s64 f32 f64 c64 c128))
            (make-typevar s '())
            (error "invalid type specifier" s)))
      (((or 'union 'struct 'typename) (? symbol? name))
        (make-typevar (list (car s) name) '()))
      ((? list? s)
        (match (memq '-> s)
          ((_ rt)
            (typespec `(procedure ,(take s (- (length s) 2)) ,rt)))
          (else (error "invalid type specifier" s))))
      (_ (error "invalid type specifier" s))))

  (define (void-type)
    (make-typevar 'void '()))

  (define (procedure-type/optionals pt1 argtypes rtype)
    (match (typevar-type pt1)
      (('procedure ptargs _)
       (call/cc
         (lambda (fail)
           `(procedure
              ,(let rec ((ptargs ptargs) (argtypes argtypes) (o #f))
                 (match ptargs
                   (('#!optional . ptargs)
                    `(#!optional ,@(rec ptargs argtypes #t)))
                   ((t more ...)
                    (if (null? argtypes)
                        (if o
                            (cons t more)
                            (fail #f))
                        (cons (car argtypes) (rec more (cdr argtypes) o))))
                   (() (if (null? argtypes)
                   	   '()
                           (fail #f)))))
              ,rtype))))
      (_ #f)))

  (define (procedure-type/argc argc argtypes rtype)
    `(procedure
       ,(let rec ((argtypes argtypes) (n 0))
          (cond ((null? argtypes) '())
          	((>= n argc) `(#!optional ,@argtypes))
          	(else (cons (car argtypes) (rec (cdr argtypes) (add1 n))))))
       ,rtype))

  (define (procedure-type? tv)
    (match (typevar-type tv)
      (('procedure _ _) #t)
      (_ #f)))

  (define (procedure-argument-types pt)
    (match (typevar-type pt)
      (('procedure at _) at)
      (_ #f)))

  (define (procedure-result-type pt)
    (match (typevar-type pt)
      (('procedure _ rt) rt)
      (_ #f)))

  (define (procedure-required-arguments pt)
    (match (typevar-type pt)
      (('procedure args _)
        (length (take-while (lambda (t) (not (eq? '#!optional t))) args)))))

  (define (refcounted-type? tv)
    (match (typevar-type tv)
      ((or 'port 'string ('numvector _) ('vector _)) #t)
      (('values t) (refcounted-type? t))
      (_ #f)))

  (define (vector-element-type tv)
    (match (typevar-type tv)
      ('vector (make-typevar '? '()))
      (('vector t) t)))

  (define (numeric-type? tv)
    (match (typevar-type tv)
      ((or 'integer 'float 'complex) #t)
      (_ #f)))

  (define (uninitialized-type? tv)
    (match (typevar-type tv)
      (('struct _) #t)
      (('union _) #t)
      (('typename _) #t)
      (_ #f)))

  (define (literal-type x node-type)
    (cond ((number? x)
            (cond ((inexact? x) 'float)
                  ((integer? x) 'integer)
                  (else 'complex)))
          ((boolean? x) 'boolean)
          ((char? x) 'char)
          ((string? x) 'string)
          ((vector? x)
            (let ((len (vector-length x)))
              (let loop ((i 0) (t '?))
                (if (>= i len)
                    `(vector ,(make-typevar t '()))
                    (match (list t (literal-type (vector-ref x i) node-type))
                      (('integer 'float)
                        (loop (add1 i) 'float))
                      (((or 'integer 'float) 'complex)
                        (loop (add1 i) 'complex))
                      (('? t) (loop (add1 i) t))
                      ((#f t) (loop (add1 i) t))
                      ((t #f) (loop (add1 i) t))
                      ((t1 t2)
                      	(if (or (equal? t1 t2)
                      		(and (procedure? t1) (procedure? t2)
                      		     (extended-procedure? t1) (extended-procedure? t2)
                      		     (match (list (procedure-data t1) (procedure-data t2))
                      		       ((('crunch:lambda n1) ('crunch:lambda n2))
                      		         (equal? (node-type n1) (node-type n2)))
                      		       (_ #f))))
                      	    (loop (add1 i) t1)
             		    `(error "vector literal `~s' has non-homogenuous type"
             		    	    ,x))))))))
          ((bytevector? x) '(numvector u8))
          ((number-vector? x) (list 'numvector (##sys#slot x 0)))
          ((and (procedure? x)
                (extended-procedure? x)
                (match (procedure-data x)
                  (('crunch:lambda n) n)
                  (_ #f))) =>
            (lambda (n)
              (let ((t (node-type n)))
                (and t
                    `(procedure ,(procedure-argument-types t) ,(procedure-result-type t))))))
          (else
            `(error "literal `~s' has unrepresentable type" ,x))))

  (define (type-copy! t t2)
    (typevar-type-set! t (typevar-type t2))
    ;;XXX is this overly optimistic?
    (typevar-value-set-set!
      t
      (merge-value-sets (typevar-value-set t) (typevar-value-set t2))))

  (define (unify-types! t1 t2 cell)
    (unify-types-internal t1 t2 cell)
    #;(let* ((t1n (typename t1))
           (t2n (typename t2))
           (f (unify-types-internal t1 t2 cell)))
      (print `(UNIFY: ,t1n + ,t2n -> ,(typename t1) ,(typename t2) - ,f ,cell))
      f))

  (define (unify-types-internal t1 t2 cell)
    (define (changed)
      (touch! cell)
      #t)
    (if (or (equal? (typevar-type t1) (typevar-type t2))
    	    (not (typevar-type t1))
    	    (not (typevar-type t2)))
        (let ((vs (merge-value-sets (typevar-value-set t1) (typevar-value-set t2))))
          (typevar-value-set-set! t1 vs)
          (typevar-value-set-set! t2 vs)
          #t)
        (let repeat ((t1 t1) (t2 t2) (f #f))
          (match (list (typevar-type t1) (typevar-type t2))
            ((#f _) #t)
            ((_ #f) #t)
            (('float 'integer)
              (type-copy! t2 t1)
              (changed))
            (('complex (or 'integer 'float))
              (type-copy! t2 t1)
              (changed))
            ((('procedure argtypes1 rtype1) ('procedure argtypes2 rtype2))
              (and-let* ((at3 (unify-argtypes! argtypes1 argtypes2 cell))
                         (f (unify-types! rtype1 rtype2 cell)))
        	(let ((vs (merge-value-sets (typevar-value-set t1)
        			            (typevar-value-set t2))))
          	  (typevar-value-set-set! t1 vs)
          	  (typevar-value-set-set! t2 vs)
          	  #t)))
            ((('pointer pt1) ('pointer pt2))
              (unify-types! pt1 pt2 cell))
            ((('vector pt1) ('vector pt2))
              (unify-types! pt1 pt2 cell))
            (('(numvector f64) '(pointer double)) #t)
            (('(numvector c128) '(pointer complex)) #t)
            ((('values t1) ('values t2))
              (unify-types! t1 t2 cell))
            (('(vector float) '(numvector f64)) #t)
            (('(vector complex) '(numvector c128)) #t)
            (('? _)
              (type-copy! t1 t2)
              (changed))
            ((_ '?)
              (type-copy! t2 t1)
              (changed))
            (_ (cond (f #f)
                     (else (repeat t2 t1 #t))))))))

  (define (unify-argtypes! at1 at2 cell)
    (call/cc
      (lambda (fail)
        (let rec ((at1 at1) (at2 at2))
          (match (list at1 at2)
            ((() ()) #t)
            ((('#!optional at1 ...) ('#!optional at2 ...))
              (rec at1 at2))
            ((('#!optional . _) _) (fail #f))
            ((_ ('#!optional . _)) (fail #f))
            (((t1 . at1) (t2 . at2))
              (and (match (list (typevar-type t1) (typevar-type t2))
                     ((('values t1) ('values t2))
                       (unify-types! t1 t2 cell))
                     ((('values t1) _)
                       (unify-types! t1 t2 cell))
                     ((_ ('values t2))
                       (unify-types! t1 t2 cell))
                     (_ (unify-types! t1 t2 cell)))
                   (rec at1 at2)))
            (_ (fail #f)))))))

  (define (merge-value-sets s1 s2 #!optional mark-escape)
    (match (list s1 s2)
      ((('?) ('?)) '(?))
      ((() s)
      	(when mark-escape
      	  (for-each
      	   (lambda (id)
      	     (when (symbol? id) (put! id 'crunch:escapes #t)))
      	   s))
      	'())
      ((s ())
      	(when mark-escape
          (for-each
            (lambda (id)
              (when (symbol? id) (put! id 'crunch:escapes #t)))
            s))
	'())
      (_ (delete '? (lset-union equal? s1 s2)))))

  ;; remove entries that refer to original if variant is already in value set
  ;; (the variant takes precedence)
  (define (pure-value-set tv)
    (let ((vs (typevar-value-set tv)))
      (remove
        (lambda (id)
          (and (symbol? id)
               (any (lambda (id2)
          	      (eq? id (get id2 'crunch:variant-of)))
        	    vs)))
        vs)))

  (define (merge-types1 t1 t2)
    (if (equal? (typevar-type t1) (typevar-type t2))
        (typevar-type t1)
        (let repeat ((t1 t1) (t2 t2) (f #f))
          (match (list (typevar-type t1) (typevar-type t2))
            (('float 'integer) 'float)
            (('complex (or 'int float)) 'complex)
            ((('procedure argtypes1 rtype1) ('procedure argtypes2 rtype2))
              (let ((at3 (merge-argtypes argtypes1 argtypes2)))
                (if at3
                    `(procedure ,at3 ,(merge-types rtype1 rtype2))
                    '*)))
            ((('pointer pt1) ('pointer pt2))
              (let ((pt3 (merge-types pt1 pt2)))
                (if pt3
                    `(pointer ,pt3)
                    '*)))
            ((('vector t1) ('vector t2))
              `(vector ,(merge-types t1 t2)))
            ((('values t1) ('values t2))
              `(values ,(merge-types t1 t2)))
            (('(numvector f64) '(pointer double)) '(numvector f64))
            (('(numvector c128) '(pointer complex)) '(numvector c128))
            (('(vector float) '(numvector f64)) '(numvector d64))
            (('(vector complex) '(numvector c128)) '(numvector c128))
            (('? t2) t2)
            ((t1 '?) t1)
            (('noreturn t2) t2)
            ((t1 'noreturn) t1)
            (_ (if f '* (repeat t2 t1 #t)))))))

  (define (merge-types t1 t2)
    (let ((t (merge-types1 t1 t2)))
      (if (eq? t '*)
          (make-typevar '* '())
          (make-typevar
            t
            (merge-value-sets (typevar-value-set t1) (typevar-value-set t2) #f)))))

  (define (merge-argtypes at1 at2)
    (call/cc
      (lambda (fail)
        (let rec ((at1 at1) (at2 at2))
          (match (list at1 at2)
            ((() ()) '())
            ((('#!optional at1 ...) ('#!optional at2 ...))
              (cons '#!optional (rec at1 at2)))
            ((('#!optional at1 ...) _) (fail #f))
            ((at1 ('#!optional at2 ...)) (fail #f))
            (((t1 . at1) (t2 . at2))
              (cons
                (match (list (typevar-type t1) (typevar-type t2))
                  ((('values t1) ('values t2))
                    (merge-types t1 t2))
                  ((('values t1) _)
                    (merge-types t1 t2))
                  ((_ ('values t2))
                    (merge-types t1 t2))
                  (_ (merge-types t1 t2)))
                (rec at1 at2)))
            (_ (fail #f)))))))

  (define (type=? t1 t2)
    (let ((tt1 (typevar-type t1))
          (tt2 (typevar-type t2)))
      (or (and (equal? tt1 tt2) (not (eq? tt1 '?)) (not (eq? tt2 '?)))
          (match (list tt1 tt2)
            ((('procedure at1 rt1) ('procedure at2 rt2))
              (let loop ((at1 at1) (at2 at2) (opt #f))
                (match (list at1 at2)
                  ((() ()) (type=? rt1 rt2))
                  ((('#!optional . at1) ('#!optional . at2)) (loop at1 at2 #t))
                  (((t1 . at1) (t2 . at2))
                    (and (not (eq? t1 '#!optional))
                         (not (eq? t2 '#!optional))
                         (or (and opt (eq? '? (typevar-type t1)) (eq? '? (typevar-type t2)))
                             (type=? t1 t2))
                         (loop at1 at2 opt))))))
            ((('pointer pt1) ('pointer pt2))
              (type=? pt1 pt2))
            ((('vector pt1) ('vector pt2))
              (type=? pt1 pt2))
            ((('values pt1) ('values pt2))
              (type=? pt1 pt2))
            (_ #f)))))

  (define (type>=? t1 t2)
    (let ((t3 (merge-types t1 t2)))
      (type=? t1 t3)))

  (define (clone-typevar t #!optional lmap)
    (define (copy x)
      (cond ((pair? x) (cons (copy (car x)) (copy (cdr x))))
      	    ((typevar? x) (clone-typevar x))
      	    (else x)))
    (make-typevar
      (copy (typevar-type t))
      (if lmap
      	  (adjust-cloned-ids (typevar-value-set t) lmap)
      	  (typevar-value-set t))))

  (define (adjust-cloned-ids vset lmap)
    (map (lambda (id)
      	   (cond ((assq id lmap) => cdr)
      	  	 (else id)))
      	  vset))

  (define (touch! cell)
    (vector-set! cell 0 #t))

  (define (known-type? tv)
    (not (eq? (typevar-type tv) '?)))

  (define (default-init tv)
    (match (typevar-value-set tv)
      (((? number? x)) x)
      (((? string? x)) x)
      (_ (match (typevar-type tv)
      	   ('string "")
      	   (('vector _) #()) ;XXX?
      	   (('numvector _) #u8()) ;XXX?
      	   ('port '#u8())
      	   (_ 0)))))

  (define (typevar->foreign-type tv)
    (match (typevar-type tv)
      ('boolean (values 'bool #t 'bool #t))
      ('string (values 'scheme-object "crunch_scheme_string" 'nonnull-c-pointer 'crunch#string-result))
      (('numvector s)
        (case s
          ((u8) (values 'scheme-object "crunch_scheme_u8vector" 'nonnull-c-pointer 'crunch#u8vector-result))
          ((s8) (values 'scheme-object "crunch_scheme_s8vector" 'nonnull-c-pointer 'crunch#s8vector-result))
          ((u16) (values 'scheme-object "crunch_scheme_u16vector" 'nonnull-c-pointer 'crunch#u16vector-result))
          ((s16) (values 'scheme-object "crunch_scheme_s16vector" 'nonnull-c-pointer 'crunch#s16vector-result))
          ((u32) (values 'scheme-object "crunch_scheme_u32vector" 'nonnull-c-pointer 'crunch#u32vector-result))
          ((s32) (values 'scheme-object "crunch_scheme_s32vector" 'nonnull-c-pointer 'crunch#s32vector-result))
          ((u64) (values 'scheme-object "crunch_scheme_u64vector" 'nonnull-c-pointer 'crunch#u64vector-result))
          ((s64) (values 'scheme-object "crunch_scheme_s64vector" 'nonnull-c-pointer 'crunch#s64vector-result))
          ((f32) (values 'scheme-object "crunch_scheme_f32vector" 'nonnull-c-pointer 'crunch#f32vector-result))
          ((f64) (values 'scheme-object "crunch_scheme_f64vector" 'nonnull-c-pointer 'crunch#f64vector-result))
          ((c64) (values 'scheme-object "crunch_scheme_c64vector" 'nonnull-c-pointer 'crunch#c64vector-result))
          ((c128) (values 'scheme-object "crunch_scheme_c128vector" 'c-pointer 'crunch#c128vector-result))
	  (else (error "invalid numeric vector type" (typename tv)))))
      (('vector t)
        (case (typevar-type t)
          ((integer) (values 'scheme-object "crunch_scheme_intvector" 'nonnull-c-pointer 'crunch#intvector-result))
          ((float) (values 'scheme-object "crunch_scheme_f64vector" 'nonnull-c-pointer 'crunch#f64vector-result))
          ((complex) (values 'scheme-object "crunch_scheme_c128vector" 'nonnull-c-pointer 'crunch#c128vector-result))
          (else (values #f #f #f #f))))
      (('pointer t)
        (receive (ft _ rt _) (typevar->foreign-type t)
          (values `(c-pointer ,ft) #t `(c-pointer ,rt) #t)))
      ((or 'port 'pointer) (values 'c-pointer #t 'c-pointer #t))
      (('struct s) (values `(struct ,s) #t `(struct ,s) #t))
      (('union s) (values `(union ,s) #t `(union ,s) #t))
      ((and t (or 'char 'void 'integer 'float 'complex 'scheme-object)) (values t #t t #t))
      (_ (values #f #f #f #f))))

  (define (foreign-call-wrapper tv name #!optional (rename identity) safe)
    (let ((argt (procedure-argument-types tv))
          (rt (procedure-result-type tv))
          (delegate #f))
      (define (wrap conv x)
        (cond ((string? conv) `(##core#inline ,conv ,x))
              ((symbol? conv) (list conv x))
              ((eq? #t conv) x)
              (else (error "bad conversion" conv x))))
      (receive (_ _ rft rres) (typevar->foreign-type rt)
        (unless rft
          (error "result type for foreign wrapper is invalid" name (typename rt)))
        (let* ((convs '())
               (tmps (map (lambda (at)
                            (when (eq? at '#!optional)
                              (error "optional arguments not allowed in wrapped procedure"
                                     name (typename tv)))
                            (let ((tmp (gensym 'a)))
                              (receive (aft afc _ _) (typevar->foreign-type at)
                                (when (string? afc)
                                  (set! convs (cons (list tmp at aft afc) convs)))
                                (unless aft
                                  (error "argument type for foreign wrapper is invalid"
                                         name (typename at)))
                                (when afc (set! delegate #t))
                                tmp)))
                       argt))
              (tmpf (gensym 'f)))
          (when delegate (set! delegate (gensym name)))
          `(,(rename 'begin)
             ,@(if delegate
                   `((,(rename 'foreign-declare)
                       ,(conc
                          "static " (typevar->c-type rt) " " delegate "("
                          (string-intersperse
                            (map (lambda (tmp at)
                                   (receive (aft afc _ _) (typevar->foreign-type at)
                                     (conc (if (string? afc)
                                     	       "C_word"
                                     	       (typevar->c-type at))
                                     	   " "
                                           tmp)))
                               tmps argt)
                            ", ")
                            ") {\n"
                          (apply
                              string-append
                              (map (match-lambda
                                     ((tmp at _ afc)
                                      (conc (typevar->c-type at)
                                            " " tmp "_ = "
                                            afc "(" tmp ");\n")))
                                convs))
                            (if (eq? 'void rft)
                                ""
                                (conc (typevar->c-type rt) " r = "))
                            name "("
                            (string-intersperse
                              (map (lambda (t)
                              	     (if (assq t convs)
                              	     	 (conc t "_")
                              	     	 (->string t)))
                              	   tmps)
                              ", ")
                            ");\n"
                            (apply
                              string-append
                              (map (lambda (c)
                                     (conc "crunch_unref(" (first c) "_);\n"))
                                convs))
                            (if (eq? 'void rft)
                                "}\n"
                                "return r;\n}\n"))))
                     '())
                   ,(let ((fl `(,(rename (if safe
                                             'foreign-safe-lambda
                   			     'foreign-lambda))
                                ,rft ,(or delegate name)
                                ,@(map (lambda (at)
                                         (typevar->foreign-type at))
                                    argt))))
                       (if (eq? #t rres)
                           fl
                            `(,(rename 'let) ((,tmpf ,fl))
                               (,(rename 'lambda) ,tmps
                                  ,(wrap rres (cons tmpf tmps)))))))))))

  (define (typevar->c-type tv)
    (match (typevar-type tv)
      ('boolean "int")
      ('float "double")
      ('complex "double complex")
      ('integer "long")
      ('void "void")
      ('char "uint32_t")
      ('string "crunch_string")
      ('port "crunch_port")
      ('scheme-object "C_word")
      (('box t) (string-append (typevar->c-type t) " *"))
      (('numvector s)
        (case s
          ((u8) "crunch_u8vector")
          ((s8) "crunch_s8vector")
          ((u16) "crunch_u16vector")
          ((s16) "crunch_s16vector")
          ((u32) "crunch_u32vector")
          ((s32) "crunch_s32vector")
          ((u64) "crunch_u64vector")
          ((s64) "crunch_s64vector")
          ((f32) "crunch_f32vector")
          ((f64) "crunch_f64vector")
          ((c64) "crunch_c64vector")
          ((c128) "crunch_c128vector")
	  (else (error "invalid numeric vector type" (typename tv)))))
     (('vector t)
       (case (typevar-type t)
         ((integer) "crunch_intvector")
         ((float) "crunch_f64vector")
         ((complex) "crunch_c128vector")
         ((boolean) "crunch_u8vector")
         ((char) "crunch_charvector")
         (else "crunch_vector")))
     (('procedure _ _) "crunch_function")
     (('pointer t)
       (if (eq? 'char (typevar-type t))
           "char *"
           (string-append (typevar->c-type t) " *")))
     ('pointer "void *")
     (('struct s) (string-append "struct " (->string s)))
     (('union s) (string-append "union " (->string s)))
     (('typename n) (->string n))
     (_ (error "type has no C representation" (typename tv)))))

)
